;; Generated package description from dicom.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "dicom" "1.6.0.20260821.0" "DICOM viewer - Digital Imaging & Communications in Medicine" '((emacs "29.1") (compat "31")) :commit "5db856bd206ac899ecd5378eb3cb655b310397f5" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("multimedia" "hypermedia" "files") :url "https://github.com/minad/dicom")
